package edu.berkeley.qfat;
import java.awt.*;
import java.util.*;
import java.io.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import edu.berkeley.qfat.geom.*;
import edu.wlu.cs.levy.CG.KDTree;
import edu.berkeley.qfat.bind.*;
import edu.berkeley.qfat.geom.Point;
import com.infomatiq.jsi.IntProcedure;

// EDGES RUN COUNTER-CLOCKWISE

public class Mesh implements Iterable<Mesh.T> {

    public static final float EPSILON = (float)0.0001;
    public static final Random random = new Random();

    private RTree<T>         triangles = new RTree<T>();
    private PointSet<Vertex> vertices  = new PointSet<Vertex>();

    public boolean option_wireframe    = false;
    public boolean option_errorNormals = false;
    public boolean option_selectable   = true;

    public void render(GL gl, Matrix m) { render(gl, m, false); }
    public void render(GL gl, Matrix m, boolean noColor) {
        if (option_wireframe) {
            gl.glDisable(GL.GL_LIGHTING);
            gl.glBegin(GL.GL_LINES);
            if (!noColor) gl.glColor3f(1, 1, 1);
            for (T t : this) {
                // fixme used to be .goodp
                m.times(t.e1().v1.p).glVertex(gl);
                m.times(t.e1().v2.p).glVertex(gl);
                m.times(t.e2().v1.p).glVertex(gl);
                m.times(t.e2().v2.p).glVertex(gl);
                m.times(t.e3().v1.p).glVertex(gl);
                m.times(t.e3().v2.p).glVertex(gl);
            }
            gl.glEnd();
            gl.glEnable(GL.GL_LIGHTING);
            return;
        }
        for(T t : this) {
            if (!noColor)
                gl.glColor4f((float)(0.25+(0.05*t.color)),
                             (float)(0.25+(0.05*t.color)),
                             (float)(0.75+(0.05*t.color)),
                             (float)0.3); 
            /*
            if (t.red) {
            gl.glColor4f((float)(0.75+(0.05*t.color)),
                         (float)(0.25+(0.05*t.color)),
                         (float)(0.25+(0.05*t.color)),
                         (float)0.3); 
            }
            */
            t.glTriangle(gl, m);
        }
        if (option_errorNormals)
            for(T t : this)
                for(Mesh.Vertex p : new Mesh.Vertex[] { t.v1(), t.v2(), t.v3() }) {
                    if (p.ok) {
                        gl.glBegin(GL.GL_LINES);
                        if (!noColor)
                            gl.glColor3f(1, 1, 1);
                        p.p.glVertex(gl);
                        p.p.plus(p.norm().times((float)p.error()*10)).glVertex(gl);
                        gl.glEnd();
                    }
                }
    }

    public boolean immutableVertices;
    public Mesh    error_against      = null;
    public double  error              = 0;

    public Mesh(boolean immutableVertices) { this.immutableVertices = immutableVertices; }

    public void makeVerticesImmutable() { this.immutableVertices = true; }
    public float error() { return (float)error; }

    public int size() { return vertices.size(); }
    public int numTriangles() { return triangles.size(); }
    public Iterable<Vertex> vertices() { return vertices; }
    public Iterator<T> iterator() { return triangles.iterator(); }

    public void rebindPoints() {
        // unbind all points
        for(Mesh.T t : this) {
            t.v1().unbind();
            t.v2().unbind();
            t.v3().unbind();
        }
        // ask edges to re-implement their bindings
        for(Mesh.T t : this) {
            t.e1().dobind();
            t.e2().dobind();
            t.e3().dobind();
        }
        System.out.println("rebound!");
    }

    public void transform(Matrix m) {
        ArrayList<Vertex> set = new ArrayList<Vertex>();
        for(Vertex v : vertices) set.add(v);
        for(Vertex v : set) v.transform(m.times(v.p), true, null);
        for(Vertex v : set) v.goodp = v.p;
    }

    public void rebuild() { /*vertices.rebuild();*/ }
    public Vec diagonal() { return vertices.diagonal(); }
    public Point centroid() { return vertices.centroid(); }
    public Vertex nearest(Point p) { return vertices.nearest(p); }

    /** compute the volume of the mesh */
    public float volume() {
        double total = 0;
        for(T t : this) {
            double area = t.area();
            Vec origin_to_centroid = new Vec(new Point(0, 0, 0), t.centroid());
            boolean facingAway = t.norm().dot(origin_to_centroid) > 0;
            double height = Math.abs(t.norm().dot(origin_to_centroid));
            total += ((facingAway ? 1 : -1) * area * height) / 3.0;
        }
        return (float)total;
    }


    public void subdivide() {
        for (Vertex v : vertices()) v.original = true;
        HashSet<E> edges = new HashSet<E>();
        HashSet<E> flip = new HashSet<E>();
        HashSet<T> tris = new HashSet<T>();
        int count = 0;
        for (T t : this) {
            tris.add(t);
            edges.add(t.e1());
            edges.add(t.e2());
            edges.add(t.e3());
            count++;
        }
        System.out.println("triangles="+count);
        count = 0;
        for(E e : edges) {
            if (e.destroyed || e.shattered) continue;
            e.shatter().edge = true;
            for(E ex : (Iterable<E>)e.getBoundPeers()) {
                Vertex m = nearest(ex.midpoint());
                m.edge = true;
                E e3 = ex.v1.getE(m).next;
                if (e3.v2.original)
                    flip.add(e3);
            }
        }

        int i=0;

        for(E e : flip) {
            e.flip();
            System.out.println("flip!");
            i++;
        }

        System.out.println("count="+count);

        rebindPoints();
        HashSet<Vertex> verts = new HashSet<Vertex>();
        for(Vertex v : vertices()) verts.add(v);
        for (Vertex v : verts)
            v.clearWish();
        for (Vertex v : verts) {
            if (v.edge) {
                for(E e = v.e; e!=null; e=e.pair.next==v.e?null:e.pair.next) {
                    if (e.v2.original) {
                        v.wish(e.v2);
                        v.wish(e.v2);
                        v.wish(e.v2);
                        v.wish(e.v2);
                    }
                }
                for(E e = v.e; e!=null; e=e.pair.next==v.e?null:e.pair.next) {
                    for(E e2 = e.v2.e; e2!=null; e2=e2.pair.next==e.v2.e?null:e2.pair.next) {
                        if (e2.v2.original) {
                            v.wish(e.v2);
                        }
                    }
                }
            }
        }
        for (Vertex v : verts)
            v.grantWish();
        System.out.println("-------------------------------------------------------------------");
        /*
        for (Vertex v : verts) {
            if (v.original) {
                int n=0;
                for(E e = v.e; e!=null; e=e.pair.next==v.e?null:e.pair.next) {
                    n++;
                    v.wish(e.midpoint());
                    v.wish(e.midpoint());
                    v.wish(e.next.pair.t.centroid());
                }
                v.avgWish();
                v.wishes = 3;
                for(int j=0; j<n-3; j++)
                    v.wish(v.getPoint());
            }
        }
        for (Vertex v : verts)
            v.avgWish();
        for (Vertex v : verts)
            v.grantWish();
        */
    }

    // Vertexices //////////////////////////////////////////////////////////////////////////////


    /** a vertex in the mesh */
    public final class Vertex extends HasQuadric implements Visitor, HasPoint {
        public void bindTo(Matrix bindingMatrix, HasBindingGroup other) {
            bindTo(bindingMatrix, other, EPSILON);
        }
        public float getMaxX() { return getPoint().getMaxX(); }
        public float getMinX() { return getPoint().getMinX(); }
        public float getMaxY() { return getPoint().getMaxY(); }
        public float getMinY() { return getPoint().getMinY(); }
        public float getMaxZ() { return getPoint().getMaxZ(); }
        public float getMinZ() { return getPoint().getMinZ(); }

        public Point p, goodp;
        public Point oldp;
        E e;                // some edge *leaving* this point

        public boolean original = false;
        public boolean edge = false;
        public boolean face = false;

        private int wishes = 0;
        private Point wish = Point.ZERO;
        public void clearWish() { wishes = 0; wish = Point.ZERO; }
        public void wish(HasPoint hp) {
            Point p = hp.getPoint();
            wishes++;
            wish = new Point(wish.x+p.x, wish.y+p.y, wish.z+p.z);
        }
        public void grantWish() {
            for(Vertex v : (Iterable<Vertex>)getBoundPeers()) {
                if (v==this) continue;
                if (v.wishes==0) continue;
                Point p = this.getBindingMatrix(v).times(v.wish.minus(Point.ZERO).div(v.wishes).plus(Point.ZERO));
                wish = p.minus(Point.ZERO).times(v.wishes).plus(wish);
                wishes += v.wishes;
                v.clearWish();
            }
            if (wishes==0) return;
            Vec d = wish.minus(Point.ZERO).div(wishes).plus(Point.ZERO).minus(getPoint());
            move(d, false);
            clearWish();
        }
        public void avgWish() {
            if (wishes==0) return;
            wish = wish.minus(Point.ZERO).div(wishes).plus(Point.ZERO);
            wishes = 1;
        }

        private boolean illegal = false;

        public boolean visible = false;

        public Point getPoint() { return p; }
        public float error() { return olderror; }

        private Vertex(Point p) {
            this.p = p;
            this.goodp = p;
            this.oldp = p;
            if (vertices.get(p) != null) throw new Error();
            vertices.add(this);
        }

        public void reinsert() {
            vertices.remove(this);
            vertices.add(this);
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) e.t.reinsert();
        }

        // the average of all adjacent points
        public Point recenter() {
            int count = 0;
            Vec vec = Vec.ZERO;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                vec = vec.plus(e.getOther(this).getPoint().minus(Point.ZERO));
                count++;
            }
            return Point.ZERO.plus(vec.div(count));
        }

        public float olderror = 0;
        public void setError(float nerror) {
            error -= olderror;
            olderror = nerror;
            error += olderror;
        }

        /*
        public Vertex hack(GL gl, Point mouse) {
            double dist = Double.MAX_VALUE;
            Vertex cur = null;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                Vertex v = e.getOther(this);
                double dist2 = v.getPoint().glProject(gl).distance(mouse);
                if ((cur==null || dist2 < dist) && v.visible) {
                    dist = dist2;
                    cur = v;
                }
            }
            return cur;
        }
        */

        public float averageTriangleArea() {
            int count = 0;
            float ret = 0;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                ret += e.t.area();
                count++;
            }
            return ret/count;
        }
        public float averageEdgeLength() {
            int count = 0;
            float ret = 0;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                ret += e.length();
                count++;
            }
            return ret/count;
        }

        public Matrix _recomputeFundamentalQuadric() {
            Matrix m = Matrix.ZERO;
            int count = 0;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                m = m.plus(e.t.norm().fundamentalQuadric(e.t.centroid()));
                count++;
            }
            if (count > 0) {
                m = m.plus(norm().fundamentalQuadric(this.p).times(count));
                count *= 2;
            }
            return m.times(1/(float)count);
        }

        public HasQuadric nearest() { return error_against==null ? null : error_against.vertices.nearest(p, this); }
        public void computeError() {
            if (error_against==null) return;
            if (nearest_in_other_mesh == null && nearest()==null) return;
            float nerror =
                nearest_in_other_mesh != null
                ? nearest_in_other_mesh.fundamentalQuadric().preAndPostMultiply(p)
                : nearest().fundamentalQuadric().preAndPostMultiply(p);
            if (quadric_count != 0)
                nerror = (nerror + quadric.preAndPostMultiply(p))/(quadric_count+1);

            if (!immutableVertices && quadric_count == 0) {
                //nerror = Math.max(nerror, 0.4f);
                //nerror *= 2;
            }
            //System.out.println(nerror);
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                double ang = e.dihedralAngle();
                if (ang > Math.PI) throw new Error();
                if (ang < -Math.PI) throw new Error();
                float minangle = (float)(Math.PI * 0.8);
                //nerror += ((ang / Math.PI)*(ang/Math.PI)) * e.length() * 0.05;

                //nerror += (1-e.t.quality())*0.0001;
                if (ang > minangle) nerror += (ang - minangle);

                //System.out.println(((ang / Math.PI)*(ang/Math.PI)) * 0.000001);
                /*
                if (e.t.aspect() < 0.2) {
                    nerror += (0.2-e.t.aspect()) * 10;
                }
                */
            }
            if (!immutableVertices) {
                Vertex n = (Vertex)nearest();
                float d = norm().dot(n.norm());
                if (d > 1 || d < -1) throw new Error();
                if (d >= 0) {
                    nerror *= (2.0f - d);
                } else {
                    nerror += 0.0003 * (2.0f + d);
                    nerror *= (2.0f + d);
                }
            }

            setError(nerror);
        }

        public boolean move(Vec vv, boolean ignoreProblems) {

            boolean good = true;

            //     t1' = M * t1
            //     t2' = t2.getMatrix(t1) * t1'
            //     t2' = t2.getMatrix(t1) * M * t1
            //     t1 =     t1.getMatrix(t2) * t2
            // M * t1 = M * t1.getMatrix(t2) * t2

            /*
            if (bindingGroup!=null && this != bindingGroup.getMaster()) {
                Matrix m2 = getBindingMatrix(bindingGroup.getMaster());
                Vec v2 = m2.times(vv.plus(getPoint())).minus(m2.times(getPoint()));
                return ((Vertex)bindingGroup.getMaster()).move(v2, ignoreProblems);
            }
            */

            Point op = this.p;
            Point pt = vv.plus(getPoint());
            Point pp = pt;
            pt = getBindingConstraint().getProjection(pp);
            if (pt==null) {
                System.out.println("constraint violation: " + getBindingConstraint());
                return false;
            }
            System.out.println(pt.minus(pp).mag() + " " + getBindingConstraint());

            for(Vertex v : (Iterable<Vertex>)getBoundPeers()) {
                Point pt2 = v.getBindingMatrix(this).times(pt);
                /*
                if (Math.abs( v.p.minus(pt2).mag() / pt.minus(op).mag() ) > 5)
                    throw new Error(v.p+" "+pt2+"\n"+op+" "+pt+"\n"+v.getBindingMatrix(this));
                if (Math.abs( v.p.minus(pt2).mag() / pt.minus(op).mag() ) < 1/5) throw new Error();
                */
                good &= v.transform(pt2, ignoreProblems, v.getBindingMatrix(this));
            }

            if (!good && !ignoreProblems) {
                for(Vertex v : (Iterable<Vertex>)getBoundPeers()) 
                    v.transform(v.oldp, true, null);
            }

            for(Vertex v : (Iterable<Vertex>)getBoundPeers())
                v.recomputeFundamentalQuadricIfNeighborChanged();
            for(Vertex v : (Iterable<Vertex>)getBoundPeers())
                v.reComputeErrorAround();
            ok = true;
            return good;
        }
        public boolean ok = true;

        /** does NOT update bound pairs! */
        private boolean transform(Point newp, boolean ignoreProblems, Matrix yes) {
            this.oldp = this.p;
            if (immutableVertices) throw new Error();

            unApplyQuadricToNeighbor();

            boolean illegalbefore = illegal;
            illegal = false;
            /*
            if (this.p.minus(newp).mag() > 0.1 && !ignoreProblems) {
                try {
                    throw new Exception(""+this.p.minus(newp).mag()+" "+ignoreProblems+" "+yes);
                } catch(Exception e) {
                    e.printStackTrace();
                }
                illegal = true;
            }
            */

            this.p = newp;
            reinsert();
            applyQuadricToNeighbor();

            if (!ignoreProblems) {
                checkLegality();
            }
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                e.v2.quadricStale = true;
            return !illegal || (illegalbefore && illegal);
        } 

        public void checkLegality() {
            /*
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                if (Math.abs(e.dihedralAngle()) > (Math.PI * 0.9) ||
                    Math.abs(e.next.dihedralAngle()) > (Math.PI * 0.9)) illegal = true;
                if (e.t.aspect() < 0.2) illegal = true;
            }
            */
            if (!illegal) triangles.range(oldp, this.p, (Visitor<T>)this);
        }

        public void reComputeErrorAround() {
            reComputeError();
            if (nearest_in_other_mesh != null)
                nearest_in_other_mesh.reComputeError();
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                e.v2.reComputeError();
        }

        public boolean visit(Object o) {
            if (o instanceof Vertex)
                return ((Vertex)o).e != null && ((Vertex)o).norm().dot(Vertex.this.norm()) >= 0;
            T t = (T)o;
            if (illegal) return false;
            for(E e = Vertex.this.e; e!=null; e=e.pair.next==Vertex.this.e?null:e.pair.next) {
                if (!t.has(e.v1) && !t.has(e.v2) && e.intersects(t)) { illegal = true; }
                if (e.t != null) {
                    if (!e.t.has(t.e1().v1) && !e.t.has(t.e1().v2) && t.e1().intersects(e.t)) { illegal = true; }
                    if (!e.t.has(t.e2().v1) && !e.t.has(t.e2().v2) && t.e2().intersects(e.t)) { illegal = true; }
                    if (!e.t.has(t.e3().v1) && !e.t.has(t.e3().v2) && t.e3().intersects(e.t)) { illegal = true; }
                }
            }
            return !illegal;
        }

        public E getEdge() { return e; }
        public E getFreeIncident() {
            E ret = getFreeIncident(e, e);
            if (ret != null) return ret;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                System.out.println(e + " " + e.t);
            throw new Error("unable to find free incident to " + this);
        }

        public E getFreeIncident(E start, E before) {
            for(E e = start; e!=null; e=e.pair.next==before?null:e.pair.next)
                if (e.pair.v2 == this && e.pair.t == null && e.pair.next.t == null)
                    return e.pair;
            return null;
        }

        public E getE(Point v2) {
            Vertex v = vertices.get(v2);
            if (v==null) return null;
            return getE(v);
        }
        public E getE(Vertex v2) {
            if (this.e!=null && this!=this.e.v1 && this!=this.e.v2) throw new RuntimeException();
            int i=0;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                if (e.v1 == this && e.v2 == v2) return e;
                i++;
                e.sanity();
                if (e.destroyed) throw new RuntimeException("fark " + i + " " + e.prev + " " + (e.prev.next==e) + " " + e.prev.destroyed);
            }
            return null;
        }

        private void glNormal(GL gl) {
            Vec norm = norm();
            gl.glNormal3f(norm.x, norm.y, norm.z);
        }
        public Vec norm() {
            Vec norm = new Vec(0, 0, 0);
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                if (e.t != null)
                    norm = norm.plus(e.t.norm().times((float)e.prev.angle()));
            return norm.norm();
        }

        public void bindTo(Vertex p) { bindTo(Matrix.ONE, p); }
    }


    /** [UNIQUE] an edge */
    public final class E extends HasBindingGroup implements Comparable<E> {

        public void bindTo(Matrix bindingMatrix, HasBindingGroup other) {
            bindTo(bindingMatrix, other, EPSILON);
        }

        public void sanity() {
            if (destroyed) return;
            if (pair!=null && (pair.v1!=v2 || pair.v2!=v1)) throw new RuntimeException();
            if (next!=null && next.v1!=v2) throw new RuntimeException();
            if (prev!=null && prev.v2!=v1) throw new RuntimeException();
        }

        public final Vertex v1, v2;
        T t;     // triangle to our "left"
        E prev;  // previous half-edge
        E next;  // next half-edge
        E pair;  // partner half-edge
        boolean shattered = false;


        /** the "edge normal" -- average of the normals of the adjacent faces */
        public Vec norm() {
            return
                (t==null || pair==null || pair.t==null)
                ? null
                : t.norm().plus(pair.t.norm()).norm();
        }

        public void glVertices(GL gl) {
            Point p1 = v1.p;
            Point p2 = v2.p;
            if (t != null) {
                p1 = p1.plus(t.centroid().minus(p1).times(0.1f));
                p2 = p2.plus(t.centroid().minus(p2).times(0.1f));
                p1 = p1.plus(norm().times(length() * 0.01f));
                p2 = p2.plus(norm().times(length() * 0.01f));
            }
            p1.glVertex(gl);
            p2.glVertex(gl);
            if (t==null || pair.t==null) return;

            Point atip = p2.minus(Point.ZERO).times(4).plus(p1.minus(Point.ZERO)).div(5).plus(Point.ZERO);
            atip = (t.norm().plus(pair.t.norm())).norm().times(getSegment().length() / 5).plus(atip);
            p2.glVertex(gl);
            atip.glVertex(gl);
        }

        public boolean intersects(T t) { return t.intersects(v1.p, v2.p); }

        public Segment getSegment() { return new Segment(v1.getPoint(), v2.getPoint()); }

        public void flip() {
            // FIXME: coplanarity check needed
            if (destroyed) return;
            for (E e : (Iterable<E>)getBoundPeers()) {
                if (!e.pair.isBoundTo(pair)) throw new RuntimeException("cannot flip!");
            }
            Vertex v1 = t.getOtherVertex(this);
            Vertex v2 = pair.t.getOtherVertex(pair);
            destroy();
            pair.destroy();
            T t1 = newT(v1, v2, this.v2);
            T t2 = newT(v2, v1, this.v1);
            t1.e1().sanity();
            t1.e2().sanity();
            t1.e3().sanity();
            t2.e1().sanity();
            t2.e2().sanity();
            t2.e3().sanity();

            for (E e : (Iterable<E>)getBoundPeers()) {
                if (e==this) continue;
                if (e.destroyed) continue;
                Vertex v1e = e.t.getOtherVertex(e);
                Vertex v2e = e.pair.t.getOtherVertex(e.pair);
                e.destroy();
                e.pair.destroy();
                if (v1e.getE(v2e)!=null) throw new RuntimeException();
                newT(v1e, v2e, e.v2).red = true;
                newT(v2e, v1e, e.v1).red = true;
                v2e.getE(v1e).bindTo(e.getBindingMatrix(this), v1.getE(v2));
                v1e.getE(v2e).bindTo(e.pair.getBindingMatrix(this.pair), v2.getE(v1));
            }

        }

        public void bindingGroupChanged() {
            HashSet<E> nbg = new HashSet<E>();
            for(E eother : (Iterable<E>)getBoundPeers()) nbg.add(eother);
            for(E eother : nbg) {
                if (next==null || prev==null) continue;
                if (eother.next==null || eother.prev==null) continue;

                Matrix m = getBindingMatrix(eother);
                if (next.isBoundTo(eother.pair.prev.pair) &&
                    !prev.isBoundTo(eother.pair.next.pair) &&
                    m.equalsModuloEpsilon(next.getBindingMatrix(eother.pair.prev.pair), EPSILON))
                    prev.bindEdge(next.getBindingMatrix(eother.pair.prev.pair), eother.pair.next.pair);
                if (!next.isBoundTo(eother.pair.prev.pair) &&
                    prev.isBoundTo(eother.pair.next.pair)  &&
                    m.equalsModuloEpsilon(prev.getBindingMatrix(eother.pair.next.pair), EPSILON))
                    next.bindEdge(prev.getBindingMatrix(eother.pair.next.pair), eother.pair.prev.pair);

                //if (next.isBoundTo(eother.prev) && !prev.isBoundTo(eother.next))
                //prev.bindTo(next.getBindingMatrix(eother.prev), eother.next);
                //if (!next.isBoundTo(eother.prev) && prev.isBoundTo(eother.next))
                //next.bindTo(prev.getBindingMatrix(eother.next), eother.prev);

                if (next.isBoundTo(eother.next) &&
                    !prev.isBoundTo(eother.prev) &&
                    m.equalsModuloEpsilon(next.getBindingMatrix(eother.next), EPSILON))
                    prev.bindEdge(next.getBindingMatrix(eother.next), eother.prev);
                if (!next.isBoundTo(eother.next) &&
                    prev.isBoundTo(eother.prev) &&
                    m.equalsModuloEpsilon(prev.getBindingMatrix(eother.prev), EPSILON))
                    next.bindEdge(prev.getBindingMatrix(eother.prev), eother.next);

            }

        }

        public float stretchRatio() {
            Vertex nearest = error_against.nearest(midpoint());
            float nearest_distance = midpoint().distance(nearest.p);
            float other_distance =
                (v1.p.distance(error_against.nearest(v1.p).p)+
                 v2.p.distance(error_against.nearest(v2.p).p))/2;
            return nearest_distance/other_distance;
        }
        public float comparator() {
            return length();
        }
        public int compareTo(E e) {
            return e.comparator() > comparator() ? 1 : -1;
        }
        public void bindEdge(Matrix m, E e) {
            bindEdge(e, m);
        }
        public void bindEdge(E e, Matrix m) {
            /*
            for(E e_ : (Iterable<E>)e.getBoundPeers()) {
                if (e.v1.getPoint().distance((e.getBindingMatrix(e_).times(e_.v1.getPoint()))) > 0.01f)
                    throw new RuntimeException("blah! " + e.v1.getPoint() + " " + e.getBindingMatrix(e_).times(e_.v1.getPoint()));
                if (e.v2.getPoint().distance((e.getBindingMatrix(e_).times(e_.v2.getPoint()))) > 0.01f)
                    throw new RuntimeException("blah! " + e.v2.getPoint() + " " + e.getBindingMatrix(e_).times(e_.v2.getPoint()));
                if (v1.getPoint().distance(m.times(e.getBindingMatrix(e_).times(e_.v1.getPoint()))) > 0.01f)
                    throw new RuntimeException("blah! " + v1.getPoint() + " " + m.times(e_.v1.getPoint()));
                if (v2.getPoint().distance(m.times(e.getBindingMatrix(e_).times(e_.v2.getPoint()))) > 0.01f)
                    throw new RuntimeException("blah! " + v2.getPoint() + " " + m.times(e_.v2.getPoint()));
            }
            */
            this.bindTo(m, e,      EPSILON);
            this.pair.bindTo(m, e.pair, EPSILON);
        }
        
        public void dobind() {
            for(E e : (Iterable<E>)getBoundPeers()) {
                if (e==this) continue;
                v1.bindTo(getBindingMatrix(e), e.v1);
                v2.bindTo(getBindingMatrix(e), e.v2);
            }
        }

        public Vertex shatter() {
            if (shattered || destroyed) return nearest(midpoint());
            shattered = true;
            E first = null;
            E firste = null;
            E firstx = null;
            E firstq = null;
            for(E e : (Iterable<E>)getBoundPeers()) {
                E enext = e.next;
                E eprev = e.prev;
                E pnext = e.pair.next;
                E pprev = e.pair.prev;
                Point mid = e.midpoint();
                Vertex r = e.next.v2;
                Vertex l = e.pair.next.v2;
                if (!e.destroyed) {
                    e.destroy();
                    e.pair.destroy();
                    newT(r.p, e.v1.p, mid,    null, 0);
                    newT(r.p, mid,    e.v2.p, null, 0);
                    newT(l.p, mid,    e.v1.p, null, 0);
                    newT(l.p, e.v2.p, mid,    null, 0);
                }
            }
            for(E e : (Iterable<E>)getBoundPeers()) {
                Point mid = e.midpoint();
                if (first==null) {
                    first = e.v1.getE(mid);
                    firste = e;
                    firstx = e.pair;
                    firstq = e.v2.getE(mid).pair;
                    continue;
                }
                e.v1.getE(mid).          bindTo(e.getBindingMatrix(firste), first);
                e.v1.getE(mid).pair.     bindTo(e.getBindingMatrix(firste), first.pair);
                e.v2.getE(mid).pair.     bindTo(e.getBindingMatrix(firste), firstq);
                e.v2.getE(mid).pair.pair.bindTo(e.getBindingMatrix(firste), firstq.pair);
            }
            /*
            first.setConstraint(firste.getAffineConstraint());
            firstq.setConstraint(firste.getAffineConstraint());
            */
            return nearest(midpoint());
        }

        public boolean destroyed = false;
        public void destroy() {
            if (destroyed) return;
            destroyed = true;
            pair.destroyed = true;

            if (t != null) t.destroy();
            t = null;

            if (pair.t != null) pair.t.destroy();
            pair.t = null;

            if (next.t != null) next.t.destroy();
            if (prev.t != null) prev.t.destroy();
            next.t = null;
            prev.t = null;

            if (pair.next.t != null) pair.next.t.destroy();
            if (pair.prev.t != null) pair.next.t.destroy();
            pair.next.t = null;
            pair.prev.t = null;

            pair.prev.next = next;
            next.prev = pair.prev;
            prev.next = pair.next;
            pair.next.prev = prev;

            if (v1.e == this) v1.e = pair.next;
            if (pair.v1.e == pair) pair.v1.e = next;

            if (v2.e == this) throw new RuntimeException();
            if (pair.v2.e == pair) throw new RuntimeException();

            /*
            next = pair;
            prev = pair;
            pair.next = this;
            pair.prev = this;
            */

            /*
            pair.prev = null;
            pair.next = null;
            next = null;
            prev = null;
            */

            /*
            sanity();
            next.sanity();
            prev.sanity();
            pair.sanity();
            */
        }

        private void sync() {
            this.prev.next = this;
            this.next.prev = this;
            this.pair.pair = this;
            if (this.next.v1 != v2) throw new Error();
            if (this.prev.v2 != v1) throw new Error();
            if (this.v1.e == null) this.v1.e = this;
            if (!added) added = true;
        }
        private boolean added = false;

        public T makeT(int colorclass) { return t==null ? (t = new T(this, colorclass)) : t; }

        public double dihedralAngle() {
            Vec v1 = t.norm().times(-1);
            Vec v2 = pair.t.norm().times(-1);
            double prod = v1.norm().dot(v2.norm());
            prod = Math.min(1,prod);
            prod = Math.max(-1,prod);
            double ret = Math.acos(prod);
            if (Double.isNaN(ret)) throw new Error("nan! " + prod);
            return ret;
        }

        /** angle between this half-edge and the next */
        public double angle() {
            Vec v1 = next.v2.p.minus(this.v2.p);
            Vec v2 = this.v1.p.minus(this.v2.p);
            return Math.acos(v1.norm().dot(v2.norm()));
        }

        public Vertex getOther(Vertex v) {
            if (this.v1 == v) return v2;
            if (this.v2 == v) return v1;
            throw new Error();
        }

        public void makeAdjacent(E e) {
            if (this.next == e) return;
            if (v2 != e.v1) throw new Error("cannot make adjacent -- no shared vertex: " + this + " " + e);
            if (t != null || e.t != null) throw new Error("cannot make adjacent -- edges not both free " + t + " " + e.t);

            E freeIncident = v2.getFreeIncident(e, this);

            e.prev.next = freeIncident.next;
            freeIncident.next.prev = e.prev;

            freeIncident.next = this.next;
            this.next.prev = freeIncident;
            
            this.next = e;
            e.prev = this;

            sync();
            freeIncident.sync();
        }

        /** creates an isolated edge out in the middle of space */
        public E(Point v1, Point v2) {
            if (vertices.get(v1) != null) throw new Error();
            if (vertices.get(v2) != null) throw new Error();
            if (v1.equals(v2)) throw new Error("attempt to create a zero-length edge!");
            this.v1 = new Vertex(v1);
            this.v2 = new Vertex(v2);
            this.prev = this.next = this.pair = new E(this, this, this);
            this.v1.e = this;
            this.v2.e = this.pair;
            sync();
        }

        /** adds a new half-edge from prev.v2 to v2 */
        public E(E prev, Point p) {
            Vertex v2;
            v2 = vertices.get(p);
            if (v2 == null) v2 = new Vertex(p);
            this.v1 = prev.v2;
            this.v2 = v2;
            this.prev = prev;
            if (prev.destroyed) throw new RuntimeException();
            if (v2.getE(v1) != null) throw new Error();
            if (v2.e==null) {
                this.next = this.pair = new E(this, this, prev.next);
            } else {
                E q = v2.getFreeIncident();
                this.next = q.next;
                this.next.prev = this;
                E z = prev.next;
                this.prev.next = this;
                this.pair = new E(q, this, z);
            }
            if (v2.e==null) v2.e = this.pair;
            sync();
        }

        /** adds a new half-edge to the mesh with a given predecessor, successor, and pair */
        public E(E prev, E pair, E next) {
            this.v1 = prev.v2;
            this.v2 = next.v1;
            if (prev.destroyed) throw new RuntimeException();
            this.prev = prev;
            this.next = next;
            this.pair = pair;
            sync();
        }
        public Point midpoint() { return new Point((v1.p.x+v2.p.x)/2, (v1.p.y+v2.p.y)/2, (v1.p.z+v2.p.z)/2); }
        public boolean has(Vertex v) { return v==v1 || v==v2; }
        public float length() { return v1.p.minus(v2.p).mag(); }
        public String toString() { return v1+"->"+v2; }

    }

    public E makeE(Point p1, Point p2) {
        Vertex v1 = vertices.get(p1);
        Vertex v2 = vertices.get(p2);
        if (v1 != null && v2 != null) {
            E e = v1.getE(v2);
            if (e != null) return e;
            e = v2.getE(v1);
            if (e != null) return e;
        }
        if (v1 != null) return new E(v1.getFreeIncident(), p2);
        if (v2 != null) return new E(v2.getFreeIncident(), p1).pair;
        return new E(p1, p2);
    }
    public boolean coalesce = false;
    private static float round(float f) {
        return Math.round(f*1000)/1000f;
    }
    public T newT(HasPoint v1, HasPoint v2, HasPoint v3) {
        return newT(v1.getPoint(), v2.getPoint(), v3.getPoint(), null, 0);
    }
    public T newT(Point v1, Point v2, Point v3, Vec norm) { return newT(v1, v2, v3, norm, 1); }
    public T newT(Point v1, Point v2, Point v3, Vec norm, int colorclass) {
        if (coalesce) {

            for(Vertex v : vertices) { if (v1.distance(v.p) < EPSILON) { v1 = v.p; break; } }
            for(Vertex v : vertices) { if (v2.distance(v.p) < EPSILON) { v2 = v.p; break; } }
            for(Vertex v : vertices) { if (v3.distance(v.p) < EPSILON) { v3 = v.p; break; } }
            /*
            v1 = new Point(round(v1.x), round(v1.y), round(v1.z));
            v2 = new Point(round(v2.x), round(v2.y), round(v2.z));
            v3 = new Point(round(v3.x), round(v3.y), round(v3.z));
            */
        }
        if (norm != null) {
            Vec norm2 = v3.minus(v1).cross(v2.minus(v1));
            float dot = norm.dot(norm2);
            //if (Math.abs(dot) < EPointSILON) throw new Error("dot products within evertsilon of each other: "+norm+" "+norm2);
            if (dot < 0) { Point p = v1; v1=v2; v2 = p; }
        }
        E e12 = makeE(v1, v2);
        E e23 = makeE(v2, v3);
        E e31 = makeE(v3, v1);
        while(e12.next != e23 || e23.next != e31 || e31.next != e12) {
            e12.makeAdjacent(e23);
            e23.makeAdjacent(e31);
            e31.makeAdjacent(e12);
        }
        T ret = e12.makeT(colorclass);
        if (e12.t == null) throw new Error();
        if (e23.t == null) throw new Error();
        if (e31.t == null) throw new Error();
        return ret;
    }

    private int max_serial = 0;
    /** [UNIQUE] a triangle (face) */
    public final class T extends Triangle {
        public final E e1;
        public final int color;
        public final int colorclass;

        public boolean red = false;
        public boolean old = false;

        public final int serial = max_serial++;
        public boolean occluded;

        public Point shatter() {
            if (destroyed) return null;
            E e = e1();
            
            HashSet<E> forward = new HashSet<E>();
            HashSet<E> backward = new HashSet<E>();
            HashSet<E> both = new HashSet<E>();

            for(E eb : (Iterable<E>)e.getBoundPeers()) {
                if (eb==e) continue;
                if (eb.next.isBoundTo(e.next) && eb.prev.isBoundTo(e.prev)) {
                    forward.add(eb);
                    both.add(eb);
                }
                if (eb.pair.next.pair.isBoundTo(e.prev) && eb.pair.prev.pair.isBoundTo(e.next)) {
                    backward.add(eb.pair);
                    both.add(eb.pair);
                }
            }

            Vertex v1 = e.t.v1();
            Vertex v2 = e.t.v2();
            Vertex v3 = e.t.v3();
            Point c = e.t.centroid();
            E e_next = e.next;
            E e_prev = e.prev;
            e.t.destroy();
            newT(v1, v2, c);
            newT(c,  v2, v3);
            newT(v3, v1, c);

            // FIXME: forward too
            for(E ex : backward) {
                Vertex v1x = ex.t.v1();
                Vertex v2x = ex.t.v2();
                Vertex v3x = ex.t.v3();
                Point cx = ex.t.centroid();
                E ex_next = ex.next;
                E ex_prev = ex.prev;
                ex.t.destroy();
                newT(v1x, v2x, cx);
                newT(cx,  v2x, v3x);
                newT(v3x, v1x, cx);

                // FIXME: i have no idea if this is right
                e.next.bindTo(e.getBindingMatrix(ex.pair), ex.prev);
                e.prev.bindTo(e.getBindingMatrix(ex.pair), ex.next);
                e.next.pair.bindTo(e.getBindingMatrix(ex.pair), ex.prev.pair);
                e.prev.pair.bindTo(e.getBindingMatrix(ex.pair), ex.next.pair);

                e_next.next.bindTo(e_next.getBindingMatrix(ex_prev.pair), ex_prev.prev.pair);
                e_next.prev.bindTo(e_next.getBindingMatrix(ex_prev.pair), ex_prev.next.pair);

                e_prev.next.bindTo(e_prev.getBindingMatrix(ex_next.pair), ex_next.prev.pair);
                e_prev.prev.bindTo(e_prev.getBindingMatrix(ex_next.pair), ex_next.next.pair);
            }

            /*

            E first = null;
            E firste = null;
            E firstx = null;
            E firstq = null;
            for(E e : (Iterable<E>)getBoundPeers()) {
                E enext = e.next;
                E eprev = e.prev;
                E pnext = e.pair.next;
                E pprev = e.pair.prev;
                Point mid = e.midpoint();
                Vertex r = e.next.v2;
                Vertex l = e.pair.next.v2;
                if (!e.destroyed) {
                    e.destroy();
                    e.pair.destroy();
                    newT(r.p, e.v1.p, mid,    null, 0);
                    newT(r.p, mid,    e.v2.p, null, 0);
                    newT(l.p, mid,    e.v1.p, null, 0);
                    newT(l.p, e.v2.p, mid,    null, 0);
                }
            }
            for(E e : (Iterable<E>)getBoundPeers()) {
                Point mid = e.midpoint();
                if (first==null) {
                    first = e.v1.getE(mid);
                    firste = e;
                    firstx = e.pair;
                    firstq = e.v2.getE(mid).pair;
                    continue;
                }
                e.v1.getE(mid).          bindTo(e.getBindingMatrix(firste), first);
                e.v1.getE(mid).pair.     bindTo(e.getBindingMatrix(firste), first.pair);
                e.v2.getE(mid).pair.     bindTo(e.getBindingMatrix(firste), firstq);
                e.v2.getE(mid).pair.pair.bindTo(e.getBindingMatrix(firste), firstq.pair);
            }
            */
            /*
            first.setConstraint(firste.getAffineConstraint());
            firstq.setConstraint(firste.getAffineConstraint());
            */
            return null;
        }


        T(E e1, int colorclass) {
            this.e1 = e1;
            E e2 = e1.next;
            E e3 = e2.next;
            if (e1==e2 || e1==e3) throw new Error();
            if (e3.next!=e1) throw new Error();
            if (e1.t!=null || e2.t!=null || e3.t!=null) throw new Error("non-manifold surface or disagreeing normals");
            e1.t = this;
            e1.next.t = this;
            e1.next.next.t = this;

            // FIXME: check for sealed/watertight surface once construction is complete (and infer normal(s)?)

            int color = Math.abs(random.nextInt());
            while(true) {
                color = color % 4;
                if (e1().pair.t != null && color == e1().pair.t.color) { color++; continue; }
                if (e2().pair.t != null && color == e2().pair.t.color) { color++; continue; }
                if (e3().pair.t != null && color == e3().pair.t.color) { color++; continue; }
                break;
            }
            this.color = color;
            this.colorclass = colorclass;
            triangles.add(this);
        }
        public E e1() { return e1; }
        public E e2() { return e1.next; }
        public E e3() { return e1.prev; }
        public Vertex v1() { return e1.v1; }
        public Vertex v2() { return e1.v2; }
        public Vertex v3() { return e1.next.v2; }
        public Point p1() { return e1.v1.p; }
        public Point p2() { return e1.v2.p; }
        public Point p3() { return e1.next.v2.p; }
        public boolean hasE(E e) { return e1==e || e1.next==e || e1.prev==e; }
        public boolean has(Vertex v) { return v1()==v || v2()==v || v3()==v; }

        public Vertex getOtherVertex(E e) {
            if (!hasE(e)) throw new RuntimeException();
            if (!e.has(v1())) return v1();
            if (!e.has(v2())) return v2();
            if (!e.has(v3())) return v3();
            throw new RuntimeException();
        }

        public void removeFromRTree() { triangles.remove(this); }
        public void addToRTree() { triangles.insert(this); }
        public void destroy() {
            if (e1 != null) {
                e1.t = null;
                e1.next.t = null;
                e1.prev.t = null;
            }
            triangles.remove(this);
            destroyed = true;
        }
        public void reinsert() { triangles.remove(this); triangles.add(this); }

        private boolean destroyed = false;
        public boolean destroyed() { return destroyed; }

        public boolean shouldBeDrawn() {
            if (e1().bindingGroupUnconstrained()) return false;
            if (e2().bindingGroupUnconstrained()) return false;
            if (e3().bindingGroupUnconstrained()) return false;
            return true;
        }

        public void glTriangle(GL gl, Matrix m) {
            gl.glPushName(serial);
            gl.glBegin(GL.GL_TRIANGLES);
            glVertices(gl, m);
            gl.glEnd();
            gl.glPopName();
        }

        /** issue gl.glVertex() for each of the triangle's points */
        public void glVertices(GL gl, Matrix m) {
            if (!shouldBeDrawn()) return;
            super.glVertices(gl, m);
        }
    }

    // Dump /////////////////////////////////////////////////////////////////////////////

    public void dump(OutputStream os) throws IOException {
        PrintWriter pw = new PrintWriter(new OutputStreamWriter(os));
        pw.println("solid dump");
        for(Mesh.T t : this) {
            Vec normal = t.norm();
            pw.println("facet normal " + normal.x + " " + normal.y + " " + normal.z);
            pw.println("  outer loop");
            for(Mesh.Vertex v : new Mesh.Vertex[] { t.v1(), t.v2(), t.v3() }) {
                pw.println("    vertex " + v.p.x + " " + v.p.y + " " + v.p.z);
            }
            pw.println("  endloop");
            pw.println("endfacet");
        }
        pw.println("endsolid dump");
        pw.flush();
    }

}
